import logging

from ..blocks.utils import plausible_timestamps
from ..signal import Signal

try:
    from PySide6 import QtWidgets

    from .widgets.plot_standalone import PlotWindow

    QT = True
except ImportError:
    QT = False


logger = logging.getLogger("asammdf")


def plot(signals: Signal | list[Signal], title: str = "", validate: bool = True, index_only: bool = False) -> None:
    """Create a stand-alone plot using the input signal or signals.

    Arguments
    ---------
    signals : iterable | Signal
        Signals to plot.
    title : str, default ''
        Window title.
    validate : bool, default True
        Consider the invalidation bits.
    """

    if QT:
        app = QtWidgets.QApplication([])
        app.setOrganizationName("py-asammdf")
        app.setOrganizationDomain("py-asammdf")
        app.setApplicationName("py-asammdf")

        if validate:
            if isinstance(signals, (tuple, list)):
                signals = [signal.validate() for signal in signals]
            else:
                signals = [signals.validate()]

        for signal in signals:
            all_ok, idx = plausible_timestamps(signal.timestamps, -1e6, 1e9)
            if not all_ok:
                signal.samples = signal.samples[idx]
                signal.timestamps = signal.timestamps[idx]

        main = PlotWindow({sig.name: sig for sig in signals})
        if title.strip():
            main.setWindowTitle(title.strip())
        else:
            if isinstance(signals, (tuple, list)):
                main.setWindowTitle(", ".join(sig.name for sig in signals))
            else:
                main.setWindowTitle(signals.name)

        app.setStyle(QtWidgets.QStyleFactory.create("Fusion"))

        app.exec_()

    else:
        raise Exception("Signal plotting requires pyqtgraph or matplotlib")



import json
import os
from pathlib import Path

from natsort import natsorted
from PySide6 import QtCore, QtGui, QtWidgets

from ..dialogs.error_dialog import ErrorDialog
from ..dialogs.messagebox import MessageBox
from ..ui.functions_manager import Ui_FunctionsManager
from ..utils import (
    check_generated_function,
    generate_python_function,
    generate_python_function_globals,
    generate_python_variables,
)
from .python_highlighter import PythonHighlighter


class FunctionsManager(Ui_FunctionsManager, QtWidgets.QWidget):
    def __init__(self, definitions, channels=None, selected_definition="", global_variables="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.original_globals = global_variables

        self.channels = channels or {}

        definitions.pop("__global_variables__", None)

        for info in definitions.values():
            info["original_definition"] = info["current_definition"] = info["definition"] = info["definition"].replace(
                "\t", "    "
            )

        self.definitions = definitions

        self.globals_definition.setPlaceholderText(
            """The global variables definition is written as Python code.

Here is a minimalistic example:

PI = 3.1415 # float value
magic_number =  9  # integer value
initial_samples = [3, 5, 6]  # list of integers
"""
        )
        self.globals_definition.setPlainText(global_variables)
        self.globals_definition.setTabStopDistance(
            QtGui.QFontMetricsF(self.globals_definition.font()).horizontalAdvance(" ") * 4
        )

        p = self.globals_definition.palette()
        for active in (QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorGroup.Inactive):
            p.setColor(active, QtGui.QPalette.ColorRole.Base, QtGui.QColor("#131314"))
            p.setColor(active, QtGui.QPalette.ColorRole.WindowText, QtGui.QColor("#ffffff"))
            p.setColor(active, QtGui.QPalette.ColorRole.Text, QtGui.QColor("#ffffff"))
        self.globals_definition.setPalette(p)
        self.globals_highlighter = PythonHighlighter(self.globals_definition.document())

        self.function_definition.setPlaceholderText(
            """The virtual channel definition is written as a Python function.
Use the 'return' statement to return a value, otherwise 'None' will automatically be returned by the function.
The last function argument must be 't=0'.

Here is a minimalistic example:

def MyAverage(main_clock=0, p_FL=0, p_FR=0, p_RL=0, p_RR=0, vehicle_speed=0, t=0): 
    if main_clock > 5000:
        return 0
    else:
        avg = (p_FL + p_FR + p_RL + p_RR) / 4
        
        if avg > 20.5 and vehicle_speed < 100:
            return avg
        else:
            return avg + 9
"""
        )
        self.function_definition.setTabStopDistance(
            QtGui.QFontMetricsF(self.function_definition.font()).horizontalAdvance(" ") * 4
        )

        p = self.function_definition.palette()
        for active in (QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorGroup.Inactive):
            p.setColor(active, QtGui.QPalette.ColorRole.Base, QtGui.QColor("#131314"))
            p.setColor(active, QtGui.QPalette.ColorRole.WindowText, QtGui.QColor("#ffffff"))
            p.setColor(active, QtGui.QPalette.ColorRole.Text, QtGui.QColor("#ffffff"))
        self.function_definition.setPalette(p)

        self.func_highlighter = PythonHighlighter(self.function_definition.document())

        self.functions_list.minimal_menu = True
        self.functions_list.all_texts = True
        self.functions_list.placeholder_text = "Press the + button to add a new function definition"
        self.functions_list.user_editable = True
        self.functions_list.setAlternatingRowColors(True)
        self.functions_list.setIconSize(QtCore.QSize(16, 16))
        self.functions_list.itemsDeleted.connect(self.definitions_deleted)

        self.add_btn.clicked.connect(self.add_definition)
        self.check_syntax_btn.clicked.connect(self.check_syntax)
        self.check_globals_syntax_btn.clicked.connect(self.check_globals_syntax)
        self.erase_btn.clicked.connect(self.erase_definitions)
        self.export_btn.clicked.connect(self.export_definitions)
        self.import_btn.clicked.connect(self.import_definitions)
        self.store_btn.clicked.connect(self.store_definition)
        self.load_original_function_btn.clicked.connect(self.load_original_function_definition)
        self.load_original_globals_btn.clicked.connect(self.load_original_globals_definition)

        self.tabs.currentChanged.connect(self.tabs_changed)

        self.refresh_functions_list(selected_definition=selected_definition)

        self.functions_list.currentItemChanged.connect(self.definition_selection_changed)

        for button in (
            self.add_btn,
            self.check_syntax_btn,
            self.check_globals_syntax_btn,
            self.erase_btn,
            self.import_btn,
            self.export_btn,
        ):
            button.setDefault(False)
            button.setAutoDefault(False)

        self.showMaximized()

    def add_definition(self):
        if previous := self.functions_list.currentItem():
            name = previous.text()
            previous.setIcon(QtGui.QIcon())
            info = self.definitions[name]
            info["current_definition"] = self.function_definition.toPlainText().replace("\t", "    ")

            ok, _ = self.check_syntax(silent=True)
            if ok:
                if info["current_definition"] != info["definition"]:
                    previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation))
                else:
                    previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_CommandLink))
            else:
                previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxCritical))

        counter = 1
        while True:
            name = f"Function{counter}"
            if name in self.definitions:
                counter += 1
            else:
                break

        self.definitions[name] = {
            "definition": f"def {name}(t=0):\n    return 0",
            "current_definition": f"def {name}(t=0):\n    return 0",
            "uuid": os.urandom(6).hex(),
        }

        self.refresh_functions_list(selected_definition=name)

    def check_globals_syntax(self, silent=False):
        trace = generate_python_variables(self.globals_definition.toPlainText())

        if trace:
            ErrorDialog(
                title="Global variables definition check",
                message="The syntax is not correct. The following error was found",
                trace=trace,
                parent=self,
            ).exec()

        elif not silent:
            MessageBox.information(
                self,
                "Global variables definition check",
                "The global variables definition appears to be correct.",
            )

        return not bool(trace)

    def check_syntax(self, silent=False, definition=None, globals_definition=None):
        if definition is None:
            definition = self.function_definition.toPlainText().replace("\t", "    ")

        if globals_definition is None:
            globals_definition = self.globals_definition.toPlainText()

        _globals = generate_python_function_globals()

        generate_python_variables(globals_definition, in_globals=_globals)

        for info in self.definitions.values():
            generate_python_function(info["definition"], in_globals=_globals)

        function_source = definition
        func, trace = generate_python_function(function_source, in_globals=_globals)

        return check_generated_function(func, trace, function_source, silent, parent=self)

    def definitions_deleted(self, deleted):
        count = self.functions_list.count()
        names = {self.functions_list.item(row).text() for row in range(count)}

        deleted = [name for name in self.definitions if name not in names]

        for name in deleted:
            del self.definitions[name]

        if not self.definitions:
            self.function_definition.setPlainText("")

    def definition_selection_changed(self, current, previous):
        if previous:
            name = previous.text()
            previous.setIcon(QtGui.QIcon())
            info = self.definitions[name]
            info["current_definition"] = self.function_definition.toPlainText().replace("\t", "    ")

            ok, _ = self.check_syntax(silent=True)
            if ok:
                if info["current_definition"] != info["definition"]:
                    previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation))
                else:
                    previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_CommandLink))
            else:
                previous.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxCritical))

        if current:
            name = current.text()
            info = self.definitions[name]
            self.function_definition.setPlainText(info["current_definition"])

            ok, _ = self.check_syntax(silent=True)
            if ok:
                if info["current_definition"] != info["definition"]:
                    current.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation))
                else:
                    current.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_CommandLink))
            else:
                current.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxCritical))

    def erase_definitions(self):
        self.functions_list.clear()
        self.definitions = {}
        self.function_definition.setPlainText("")

    def export_definitions(self, *args):
        self.refresh_definitions()

        for name, info in self.definitions.items():
            if info["current_definition"] != info["definition"]:
                result = MessageBox.question(
                    self, "Unsaved function definitions", "Do you want to review the functions before exporting them?"
                )
                if result == MessageBox.StandardButton.No:
                    break
                else:
                    for row in range(self.functions_list.count()):
                        item = self.functions_list.item(row)
                        if item.text() == name:
                            self.functions_list.setCurrentRow(row)
                            self.functions_list.scrollToItem(item)
                            break
                    return

        file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Select functions definition export file",
            "",
            "Functions definition file (*.def)",
        )

        if file_name and Path(file_name).suffix.lower() == ".def":
            definitions = {name: info["current_definition"] for name, info in self.definitions.items()}
            definitions["__global_variables__"] = self.globals_definition.toPlainText().replace("\t", "    ")
            Path(file_name).write_text(json.dumps(definitions, indent=2))

    def import_definitions(self, *args):
        self.refresh_definitions()

        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Select functions definition file",
            "",
            "Functions definitions file (*.def)",
            "Functions definitions file (*.def)",
        )

        if file_name and Path(file_name).suffix.lower() == ".def":
            file_name = Path(file_name)

            with open(file_name) as infile:
                info = json.load(infile)

            self.globals_definition.setPlainText(info.pop("__global_variables__", "").replace("\t", "    "))

            for name, definition in info.items():
                definition = definition.replace("\t", "    ")

                if name in self.definitions:
                    self.definitions[name]["definition"] = self.definitions[name]["current_definition"] = definition
                else:
                    self.definitions[name] = {
                        "definition": definition,
                        "current_definition": definition,
                        "uuid": os.urandom(6).hex(),
                    }

            self.refresh_functions_list()

    def load_original_globals_definition(self):
        self.globals_definition.setPlainText(self.original_globals)

    def load_original_function_definition(self):
        item = self.functions_list.currentItem()
        if not item:
            return

        current_name = item.text()
        info = self.definitions.pop(current_name)
        original_definition = info["original_definition"]
        info["current_definition"] = original_definition

        self.function_definition.setPlainText(original_definition)

        self.definitions[current_name] = info

        self.refresh_functions_list()

    def refresh_definitions(self):
        _globals = generate_python_function_globals()

        generate_python_variables(self.globals_definition.toPlainText(), in_globals=_globals)

        for info in self.definitions.values():
            generate_python_function(info["current_definition"], in_globals=_globals)

        function_source = self.function_definition.toPlainText().replace("\t", "    ")
        func, trace = generate_python_function(function_source, in_globals=_globals)

        if func is not None:
            name = func.__name__
            if name in self.definitions:
                self.definitions[name]["current_definition"] = self.function_definition.toPlainText().replace(
                    "\t", "    "
                )

    def refresh_functions_list(self, selected_definition=""):
        self.functions_list.blockSignals(True)
        self.functions_list.clear()

        if self.definitions:
            items = []
            names = natsorted(self.definitions)

            for name in names:
                info = self.definitions[name]
                ok, func = self.check_syntax(silent=True, definition=info["definition"])

                if ok:
                    if info["current_definition"] != info["definition"]:
                        items.append(
                            QtWidgets.QListWidgetItem(
                                self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation),
                                name,
                                self.functions_list,
                            )
                        )
                    else:
                        items.append(
                            QtWidgets.QListWidgetItem(
                                self.style().standardIcon(QtWidgets.QStyle.SP_CommandLink),
                                name,
                                self.functions_list,
                            )
                        )
                else:
                    items.append(
                        QtWidgets.QListWidgetItem(
                            self.style().standardIcon(QtWidgets.QStyle.SP_MessageBoxCritical),
                            name,
                            self.functions_list,
                        )
                    )

            self.functions_list.blockSignals(False)

            if selected_definition:
                if selected_definition in names:
                    self.functions_list.setCurrentRow(names.index(selected_definition))
                elif names:
                    self.functions_list.setCurrentRow(0)
        else:
            self.functions_list.blockSignals(False)

    def store_definition(self, *args):
        item = self.functions_list.currentItem()
        if not item:
            return

        current_name = item.text()
        info = self.definitions.pop(current_name)
        info["definition"] = info["current_definition"] = self.function_definition.toPlainText().replace("\t", "    ")

        ok, func = self.check_syntax(silent=True)
        if ok:
            func_name = func.__name__

            if current_name != func_name and func_name in self.definitions:
                MessageBox.information(
                    self,
                    "Invalid function name",
                    f'The name "{func_name}" is already given to another function.\nThe function names must be unique',
                )

            else:
                current_name = func_name

        self.definitions[current_name] = info

        self.refresh_functions_list(selected_definition=current_name)

    def tabs_changed(self, index):
        self.check_globals_syntax(silent=True)



#!/usr/bin/env python

"""PySide6 QPlainTextEdit syntax highlight for Python source code.
The original code if found here
https://wiki.python.org/moin/PyQt/Python%20syntax%20highlighting
"""

import builtins
import keyword

from PySide6 import QtCore, QtGui


def format(color, style=""):
    """Return a QTextCharFormat with the given attributes."""
    _color = QtGui.QColor(color)

    _format = QtGui.QTextCharFormat()
    _format.setForeground(_color)
    if "bold" in style:
        _format.setFontWeight(QtGui.QFont.Weight.Bold)
    if "italic" in style:
        _format.setFontItalic(True)

    _format.setBackground(QtGui.QBrush(QtGui.QColor("#131314")))

    return _format


# Syntax styles that can be shared by all languages
STYLES = {
    "keyword": format("#ec6529"),
    "operator": format("#ff0000"),
    "brace": format("#ff0000"),
    "defclassname": format("#9e5fdd", "bold"),
    "defclass": format("#ec8549", "bold"),
    "string": format("#2f8f3d"),
    "string2": format("#2f8f3d"),
    "comment": format("#65c2e5", "italic"),
    "self": format("#93548b", "italic"),
    "numbers": format("#33ccff"),
    "builtins": format("#567ac5"),
}


class PythonHighlighter(QtGui.QSyntaxHighlighter):
    """Syntax highlighter for the Python language."""

    # Python keywords
    keywords = keyword.kwlist + getattr(keyword, "softkwlist", [])

    builtins = dir(builtins)

    # Python operators
    operators = [
        r"=",
        # Comparison
        r"==",
        "!=",
        "<",
        "<=",
        ">",
        ">=",
        # Arithmetic
        r"\+",
        "-",
        r"\*",
        "/",
        "//",
        r"\%",
        r"\*\*",
        # In-place
        r"\+=",
        "-=",
        r"\*=",
        "/=",
        r"\%=",
        # Bitwise
        r"\^",
        r"\|",
        r"\&",
        r"\~",
        ">>",
        "<<",
    ]

    # Python braces
    braces = [
        r"\{",
        r"\}",
        r"\(",
        r"\)",
        r"\[",
        r"\]",
    ]

    def __init__(self, parent: QtGui.QTextDocument) -> None:
        super().__init__(parent)

        # Multi-line strings (expression, flag, style)
        self.tri_single = (QtCore.QRegularExpression("'''"), 1, STYLES["string2"])
        self.tri_double = (QtCore.QRegularExpression('"""'), 2, STYLES["string2"])

        rules = []

        # Keyword, operator, and brace rules
        rules += [(rf"\b{w}\b", 0, STYLES["keyword"]) for w in PythonHighlighter.keywords]
        rules += [(o, 0, STYLES["operator"]) for o in PythonHighlighter.operators]
        rules += [(b, 0, STYLES["brace"]) for b in PythonHighlighter.braces]
        rules += [(rf"\b{b}\b", 0, STYLES["builtins"]) for b in PythonHighlighter.builtins]

        # All other rules
        rules += [
            # 'self'
            (r"\bself\b", 0, STYLES["self"]),
            # 'def' followed by an identifier
            (r"\bdef\b\s*(\w+)", 0, STYLES["defclassname"]),
            (r"\bdef\b", 0, STYLES["defclass"]),
            # 'class' followed by an identifier
            (r"\bclass\b\s*(\w+)", 0, STYLES["defclassname"]),
            (r"\bclass\b", 0, STYLES["defclass"]),
            # Numeric literals
            (r"\b[+-]?[0-9]+\b", 0, STYLES["numbers"]),
            (r"\b[+-]?0[xX][0-9A-Fa-f]+\b", 0, STYLES["numbers"]),
            (r"\b[+-]?[0-9]+(\.[0-9]+)\b", 0, STYLES["numbers"]),
            (r"\b[+-]?[0-9]+([eE][+-]?[0-9]+)\b", 0, STYLES["numbers"]),
            (r"\b[+-]?[0-9]+(\.[0-9]+)([eE][+-]?[0-9]+)\b", 0, STYLES["numbers"]),
            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, STYLES["string"]),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, STYLES["string"]),
            # From '#' until a newline
            (r"#[^\n]*", 0, STYLES["comment"]),
        ]

        # Build a QRegularExpression for each pattern
        self.rules = [(QtCore.QRegularExpression(pat), index, fmt) for (pat, index, fmt) in rules]

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text."""

        self.tripleQuoutesWithinStrings = []
        # Do other syntax formatting
        for expression, nth, format in self.rules:
            iterator = expression.globalMatch(text)

            while iterator.hasNext():
                match = iterator.next()

                length = match.capturedEnd() - match.capturedStart()
                index = match.capturedStart()
                self.setFormat(index, length, format)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)

    def match_multiline(self, text, delimiter, in_state, style):
        """Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QRegularExpression`` for triple-single-quotes or triple-double-quotes,
        and ``in_state`` should be a unique integer to represent the
        corresponding state changes when inside those strings. Returns True if
        we're still inside a multi-line string when this function is finished.
        """
        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            match = delimiter.match(text)
            start = match.capturedStart()
            # skipping triple quotes within strings
            if start in self.tripleQuoutesWithinStrings:
                return False
            # Move past this match
            add = match.capturedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            match = delimiter.match(text, start + add)
            end = match.capturedStart()
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + match.capturedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            match = delimiter.match(text, start + length)
            start = match.capturedStart()

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False



import json
from struct import pack

from natsort import natsorted
from PySide6 import QtCore, QtGui, QtWidgets

from ..serde import extract_mime_names


class ListWidget(QtWidgets.QListWidget):
    itemsDeleted = QtCore.Signal(list)
    set_time_offset = QtCore.Signal(list)
    items_rearranged = QtCore.Signal()
    add_channels_request = QtCore.Signal(list)
    show_properties = QtCore.Signal(object)
    insert_computation = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.details_enabled = False

        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.InternalMove)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_menu)

        self.setAlternatingRowColors(True)

        self.can_delete_items = True
        self.setAcceptDrops(True)

        self.itemSelectionChanged.connect(self.item_selection_changed)

        self.show()
        self._has_hidden_items = False

    def item_selection_changed(self, item=None):
        selection = list(self.selectedItems())
        for row in range(self.count()):
            item = self.item(row)
            if item in selection:
                widget = self.itemWidget(item)
                if widget is not None:
                    widget.set_selected(True)
            else:
                widget = self.itemWidget(item)
                if widget is not None:
                    widget.set_selected(False)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if key == QtCore.Qt.Key.Key_Delete and self.can_delete_items:
            selected_items = self.selectedItems()
            deleted = []
            for item in selected_items:
                row = self.row(item)
                item_widget = self.itemWidget(item)
                deleted.append(getattr(item_widget, "uuid", None))
                if hasattr(item_widget, "disconnect_slots"):
                    item_widget.disconnect_slots()
                self.takeItem(row)
            if deleted:
                self.itemsDeleted.emit(deleted)
            event.accept()

        elif key == QtCore.Qt.Key.Key_Space and modifiers == QtCore.Qt.KeyboardModifier.NoModifier:
            event.accept()
            selected_items = self.selectedItems()
            if not selected_items:
                return

            states = [self.itemWidget(item).display.checkState() for item in selected_items]

            if any(state == QtCore.Qt.CheckState.Unchecked for state in states):
                state = QtCore.Qt.CheckState.Checked
            else:
                state = QtCore.Qt.CheckState.Unchecked
            for item in selected_items:
                wid = self.itemWidget(item)
                wid.display.setCheckState(state)

        elif key == QtCore.Qt.Key.Key_Space and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            event.accept()
            selected_items = self.selectedItems()
            if not selected_items:
                return

            states = [self.itemWidget(item).individual_axis.checkState() for item in selected_items]

            if any(state == QtCore.Qt.CheckState.Unchecked for state in states):
                state = QtCore.Qt.CheckState.Checked
            else:
                state = QtCore.Qt.CheckState.Unchecked
            for item in selected_items:
                wid = self.itemWidget(item)
                wid.individual_axis.setCheckState(state)

        elif modifiers == QtCore.Qt.KeyboardModifier.ControlModifier and key == QtCore.Qt.Key.Key_C:
            selected_items = self.selectedItems()
            if not selected_items:
                event.accept()
                return

            self.itemWidget(selected_items[0]).keyPressEvent(event)

        elif modifiers == (
            QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier
        ) and key in (
            QtCore.Qt.Key.Key_C,
            QtCore.Qt.Key.Key_P,
        ):
            selected_items = self.selectedItems()
            if not selected_items:
                event.accept()
                return
            self.itemWidget(selected_items[0]).keyPressEvent(event)

        else:
            super().keyPressEvent(event)

    def startDrag(self, supportedActions):
        selected_items = self.selectedItems()

        mimeData = QtCore.QMimeData()

        data = []

        for item in selected_items:
            entry = item.entry
            computation = item.computation

            widget = self.itemWidget(item)

            color = widget.color
            unit = widget.unit

            if entry == (-1, -1):
                info = {
                    "name": item.name,
                    "computation": computation,
                    "computed": True,
                    "unit": unit,
                    "color": color,
                }
                info = json.dumps(info).encode("utf-8")
            else:
                info = item.name.encode("utf-8")

            data.append(
                pack(
                    f"<12s3q{len(info)}s",
                    str(item.origin_uuid).encode("ascii"),
                    entry[0],
                    entry[1],
                    len(info),
                    info,
                )
            )

        mimeData.setData("application/octet-stream-asammdf", QtCore.QByteArray(b"".join(data)))

        drag = QtGui.QDrag(self)
        drag.setMimeData(mimeData)
        drag.exec(QtCore.Qt.DropAction.CopyAction)

    def dragEnterEvent(self, e):
        if e.mimeData().hasFormat("application/octet-stream-asammdf"):
            e.accept()
        super().dragEnterEvent(e)

    def dropEvent(self, e):
        if e.source() is self:
            super().dropEvent(e)
            self.items_rearranged.emit()
        else:
            data = e.mimeData()
            if data.hasFormat("application/octet-stream-asammdf"):
                names = extract_mime_names(data)
                self.add_channels_request.emit(names)
            else:
                super().dropEvent(e)

    def open_menu(self, position):
        item = self.itemAt(position)
        if item is None:
            return

        else:
            menu = QtWidgets.QMenu()
            menu.addAction(f"{self.count()} items in the list")
            menu.addSeparator()
            menu.addAction("Copy name (Ctrl+C)")
            menu.addAction("Copy display properties (Ctrl+Shift+C)")
            menu.addAction("Paste display properties (Ctrl+Shift+P)")
            menu.addSeparator()
            menu.addAction("Enable all")
            menu.addAction("Disable all")
            menu.addAction("Enable all but this")
            menu.addSeparator()
            if self._has_hidden_items:
                show_hide = "Show disabled items"
            else:
                show_hide = "Hide disabled items"
            menu.addAction(show_hide)
            menu.addSeparator()

            menu.addAction("Add to common Y axis")
            menu.addAction("Remove from common Y axis")
            menu.addSeparator()
            menu.addAction("Set unit")
            menu.addAction("Set precision")
            menu.addSeparator()
            menu.addAction("Relative time base shift")
            menu.addAction("Set time base start offset")
            menu.addSeparator()
            menu.addAction("Insert computation using this channel")
            menu.addSeparator()
            menu.addAction("Delete (Del)")
            menu.addSeparator()
            menu.addAction("Toggle details")
            menu.addAction("File/Computation properties")

            action = menu.exec_(self.viewport().mapToGlobal(position))

        if action is None:
            return

        if action.text() == "Copy name (Ctrl+C)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_C, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.itemWidget(item).keyPressEvent(event)

        elif action.text() == "Copy display properties (Ctrl+Shift+C)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_C,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.itemWidget(item).keyPressEvent(event)

        elif action.text() == "Paste display properties (Ctrl+Shift+P)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_P,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.itemWidget(item).keyPressEvent(event)

        elif action.text() == "Enable all":
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                widget.display.setCheckState(QtCore.Qt.CheckState.Checked)

        elif action.text() == "Disable all":
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                widget.display.setCheckState(QtCore.Qt.CheckState.Unchecked)

        elif action.text() == "Enable all but this":
            selected_items = self.selectedItems()
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                if item in selected_items:
                    widget.display.setCheckState(QtCore.Qt.CheckState.Unchecked)
                else:
                    widget.display.setCheckState(QtCore.Qt.CheckState.Checked)

        elif action.text() == show_hide:
            if self._has_hidden_items:
                for i in range(self.count()):
                    item = self.item(i)
                    item.setHidden(False)
            else:
                for i in range(self.count()):
                    item = self.item(i)
                    widget = self.itemWidget(item)
                    if not widget.display.isChecked():
                        item.setHidden(True)

            self._has_hidden_items = not self._has_hidden_items

        elif action.text() == "Add to common Y axis":
            selected_items = self.selectedItems()
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                if item in selected_items:
                    widget.ylink.setCheckState(QtCore.Qt.CheckState.Checked)

        elif action.text() == "Remove from common Y axis":
            selected_items = self.selectedItems()
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                if item in selected_items:
                    widget.ylink.setCheckState(QtCore.Qt.CheckState.Unchecked)

        elif action.text() == "Set unit":
            selected_items = self.selectedItems()

            unit, ok = QtWidgets.QInputDialog.getText(self, "Set new unit", "Unit:")

            if ok:
                selected_items = self.selectedItems()
                for i in range(self.count()):
                    item = self.item(i)
                    widget = self.itemWidget(item)
                    if item in selected_items:
                        widget.unit = unit
                        widget.update()

        elif action.text() == "Set precision":
            selected_items = self.selectedItems()

            precision, ok = QtWidgets.QInputDialog.getInt(self, "Set new precision (float decimals)", "Precision:")

            if ok and 0 <= precision <= 15:
                for i in range(self.count()):
                    item = self.item(i)
                    widget = self.itemWidget(item)
                    if item in selected_items:
                        widget.set_precision(precision)
                        widget.update()

        elif action.text() in (
            "Relative time base shift",
            "Set time base start offset",
        ):
            selected_items = self.selectedItems()
            if selected_items:
                if action.text() == "Relative time base shift":
                    offset, ok = QtWidgets.QInputDialog.getDouble(
                        self, "Relative offset [s]", "Offset [s]:", decimals=6
                    )
                    absolute = False
                else:
                    offset, ok = QtWidgets.QInputDialog.getDouble(
                        self,
                        "Absolute time start offset [s]",
                        "Offset [s]:",
                        decimals=6,
                    )
                    absolute = True
                if ok:
                    uuids = []

                    for i in range(self.count()):
                        item = self.item(i)
                        widget = self.itemWidget(item)
                        if item in selected_items:
                            uuids.append(widget.uuid)
                    self.set_time_offset.emit([absolute, offset, *uuids])

        elif action.text() == "Delete (Del)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_Delete, QtCore.Qt.KeyboardModifier.NoModifier
            )
            self.keyPressEvent(event)

        elif action.text() == "Toggle details":
            self.details_enabled = not self.details_enabled
            for i in range(self.count()):
                item = self.item(i)
                widget = self.itemWidget(item)
                widget.details.setVisible(self.details_enabled)
                item.setSizeHint(widget.sizeHint())

        elif action.text() == "File/Computation properties":
            selected_items = self.selectedItems()
            if len(selected_items) == 1:
                item = selected_items[0]
                self.show_properties.emit(self.itemWidget(item).uuid)

        elif action.text() == "Insert computation using this channel":
            selected_items = self.selectedItems()
            if len(selected_items) == 1:
                item = selected_items[0]
                self.insert_computation.emit(self.itemWidget(item)._name)


class MinimalListWidget(QtWidgets.QListWidget):
    itemsDeleted = QtCore.Signal(list)
    itemsPasted = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.InternalMove)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)

        self.setAlternatingRowColors(True)

        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_menu)

        self.setAcceptDrops(True)
        self.show()

        self.itemSelectionChanged.connect(self.item_selection_changed)

        self.minimal_menu = False
        self.all_texts = False
        self.placeholder_text = ""

        self.user_editable = True

    def item_selection_changed(self, item=None):
        try:
            selection = list(self.selectedItems())
            for row in range(self.count()):
                item = self.item(row)
                if item in selection:
                    self.itemWidget(item).set_selected(True)
                else:
                    self.itemWidget(item).set_selected(False)
        except:
            pass

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        if key == QtCore.Qt.Key.Key_Delete and self.user_editable:
            selected_items = self.selectedItems()
            deleted = []

            if self.all_texts:
                to_delete = set()
                for item in selected_items:
                    row = self.row(item)
                    deleted.append(row)
                    item_widget = self.itemWidget(item)
                    if hasattr(item_widget, "disconnect_slots"):
                        item_widget.disconnect_slots()
                    to_delete.add(item.text())

                all_texts = set()

                count = self.count()
                for row in range(count):
                    item = self.item(row)
                    all_texts.add(item.text())

                self.clear()
                self.addItems(natsorted(all_texts - to_delete))

            else:
                for item in selected_items:
                    row = self.row(item)
                    deleted.append(row)
                    item_widget = self.itemWidget(item)
                    if hasattr(item_widget, "disconnect_slots"):
                        item_widget.disconnect_slots()
                    self.takeItem(row)

            if deleted:
                self.itemsDeleted.emit(deleted)

            event.accept()

        elif key == QtCore.Qt.Key.Key_C and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            text = []
            for item in self.selectedItems():
                try:
                    text.append(self.itemWidget(item).text())
                except:
                    text.append(item.text())

            if text:
                text = "\n".join(text)
            else:
                text = ""

            QtWidgets.QApplication.instance().clipboard().setText(text)
            event.accept()

        elif (
            key == QtCore.Qt.Key.Key_V
            and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier
            and self.user_editable
        ):
            lines = QtWidgets.QApplication.instance().clipboard().text().splitlines()
            if lines:
                try:
                    self.addItems(lines)
                    self.itemsPasted.emit()
                except:
                    pass
            event.accept()
        else:
            super().keyPressEvent(event)

    def open_menu(self, position):
        menu = QtWidgets.QMenu()

        if self.minimal_menu:
            if self.count() > 0:
                menu.addAction(f"{self.count()} items in the list")
                if self.user_editable:
                    menu.addSeparator()
                    menu.addAction("Delete (Del)")
            else:
                return
        else:
            if self.count() == 0:
                menu.addAction(f"{self.count()} items in the list")
                menu.addSeparator()
                if self.user_editable:
                    menu.addAction("Paste names (Ctrl+V)")
            else:
                menu.addAction(f"{self.count()} items in the list")
                menu.addSeparator()
                menu.addAction("Copy names (Ctrl+C)")
                if self.user_editable:
                    menu.addAction("Paste names (Ctrl+V)")
                    menu.addSeparator()
                    menu.addAction("Delete (Del)")

        action = menu.exec_(self.viewport().mapToGlobal(position))

        if action is None:
            return

        if action.text() == "Delete (Del)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_Delete, QtCore.Qt.KeyboardModifier.NoModifier
            )
            self.keyPressEvent(event)
        elif action.text() == "Copy names (Ctrl+C)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_C, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)
        elif action.text() == "Paste names (Ctrl+V)":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_V, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.count() == 0 and self.placeholder_text:
            painter = QtGui.QPainter(self.viewport())
            painter.save()
            col = self.palette().placeholderText().color()
            painter.setPen(col)
            fm = self.fontMetrics()
            elided_text = fm.elidedText(
                self.placeholder_text, QtCore.Qt.TextElideMode.ElideRight, self.viewport().width()
            )
            painter.drawText(self.viewport().rect(), QtCore.Qt.AlignmentFlag.AlignCenter, elided_text)
            painter.restore()

    def search(self, attribute="", value=None):
        index = None
        count = self.count()
        for row in range(count):
            item = self.item(row)

            if hasattr(item, attribute):
                if getattr(item, attribute) == value:
                    index = row
                    break
            else:
                widget = self.itemWidget(item)

                if hasattr(widget, attribute):
                    if getattr(widget, attribute) == value:
                        index = row
                        break

        return index



#!/usr/bin/env python
import pathlib
from unittest import mock

from PySide6 import QtCore, QtTest, QtWidgets

from asammdf.gui.dialogs.channel_group_info import ChannelGroupInfoDialog
from asammdf.gui.dialogs.channel_info import ChannelInfoDialog
from asammdf.gui.widgets.numeric import Numeric
from asammdf.gui.widgets.plot import Plot
from asammdf.gui.widgets.tabular import Tabular
from test.asammdf.gui.widgets.test_BaseFileWidget import TestFileWidget

# Note: If it's possible and make sense, use self.subTests
# to avoid initializing widgets multiple times and consume time.


class TestTabChannels(TestFileWidget):
    def tearDown(self):
        if self.mc_ErrorDialog.call_args:
            if (
                "<class 'RuntimeError'>" in self.mc_ErrorDialog.call_args.kwargs["message"]
                and self.mc_ErrorDialog.call_count == 1
            ):
                self.mc_ErrorDialog.reset_mock()
        self.mc_ErrorDialog.assert_not_called()
        self.mc_widget_ed.assert_not_called()
        super().tearDown()

    def test_load_file_with_metadata(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement which has valid display file in metadata.
            - Store opened sub-windows typy, titles and channels names
            - Close all sub-windows
            - Press PushButton: "Load offline windows"
                - Simulate that dspf file from metadata was selected
        Expected:
            - Loaded sub-windows are the same as sub-windows opened with measurement file
        """

        def get_dspf_data(iterator):
            data = []
            for sub_win in iterator:
                w = sub_win.widget()
                if isinstance(w, Plot):
                    data.append(
                        {
                            "channels": [
                                w.channel_selection.topLevelItem(_).name
                                for _ in range(w.channel_selection.topLevelItemCount())
                            ],
                            "plot_bg_color": w.plot.backgroundBrush().color().name(),
                        }
                    )
                elif isinstance(w, Numeric):
                    data.append([sig.name for sig in w.channels.dataView.backend.signals])
                elif isinstance(w, Tabular):
                    data.append(list(w.tree.pgdf.df.columns))
                return data

        # Setup
        measurement_file = str(pathlib.Path(self.resource, "test_metadata.mf4"))
        # valid_dsp = str(pathlib.Path(self.resource, "valid.dsp"))

        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)

        # Prepare expected results
        dspf_file = self.widget.mdf.header._common_properties.get("pr_display_file", "")
        dspf_path = str(pathlib.Path(self.resource, dspf_file))

        sub_windows = self.get_sub_windows()
        titles = sorted(w.windowTitle() for w in self.widget.mdi_area.subWindowList())
        dspf_data = get_dspf_data(self.widget.mdi_area.subWindowList())

        # Close all sub-windows
        for w in self.widget.mdi_area.subWindowList():
            w.close()

        # Evaluate that all sub-windows was closed
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

        # Open DSPF
        with mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName:
            mo_getOpenFileName.return_value = dspf_path, None
            QtTest.QTest.mouseClick(
                self.widget.load_channel_list_btn,
                QtCore.Qt.MouseButton.LeftButton,
            )
            self.processEvents()

            # Evaluate
            self.assertListEqual(sub_windows, self.get_sub_windows())
            self.assertListEqual(titles, sorted(w.windowTitle() for w in self.widget.mdi_area.subWindowList()))
            for old, new in zip(dspf_data, get_dspf_data(self.widget.mdi_area.subWindowList()), strict=False):
                if isinstance(old, dict):
                    self.assertListEqual(new["channels"], old["channels"])
                    self.assertEqual(new["plot_bg_color"], old["plot_bg_color"])
                elif isinstance(old, list):
                    self.assertListEqual(old, new)

    def test_PushButton_LoadOfflineWindows_DSP(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Press PushButton: "Load offline windows"
                - Simulate that valid file was selected
        Expected:
            - Evaluate that Plot Window is added in "mdi_area"
        """
        # Setup
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))
        valid_dsp = str(pathlib.Path(self.resource, "valid.dsp"))
        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)

        with mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName:
            mo_getOpenFileName.return_value = valid_dsp, None
            QtTest.QTest.mouseClick(
                self.widget.load_channel_list_btn,
                QtCore.Qt.MouseButton.LeftButton,
            )
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
            widget_types = self.get_sub_windows()
            self.assertIn("Plot", widget_types)

    @mock.patch("asammdf.gui.widgets.file.ErrorDialog")
    def test_PushButton_LoadOfflineWindows_DSPF(self, mc_file_ErrorDialog):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Case 0:
                - Press PushButton: "Load offline windows"
                    - Simulate that no file was selected
            - Case 1:
                - Press PushButton: "Load offline windows"
                    - Simulate that file with non-supported extension was selected
            - Case 2:
                - Press PushButton: "Load offline windows"
                    - Simulate that invalid "dspf" file was selected: json decode error
            - Case 3:
                - Press PushButton: "Load offline windows"
                    - Simulate that invalid "dspf" file was selected: Plot Section Key Error
            - Case 4:
                - Press PushButton: "Load offline windows"
                    - Simulate that invalid "dspf" file was selected: Numeric Section Key Error
            - Case 5:
                - Press PushButton: "Load offline windows"
                    - Simulate that invalid "dspf" file was selected: Tabular Section Key Error
            - Case 6:
                - Press PushButton: "Load offline windows"
                    - Simulate that valid "dspf" file was selected
        Expected:
            - Case 0: No dpsf file was selected
                - Evaluate that method "load_window" is not called
            - Case 1: File with non-supported extension was selected
                - Evaluate that method "load_window" is not called
            - Case 2: Json Decode Error
                - Evaluate that method "load_window" is not called
                - Evaluate that "ErrorDialog" widget is called
            - Case 3: Plot Section Key Error
                - Evaluate that there is no Plot window in "mdi_area"
                - Evaluate that "ErrorDialog" widget is called
            - Case 4: Numeric Section Key Error
                - Evaluate that there is no Numeric window in "mdi_area"
                - Evaluate that "ErrorDialog" widget is called
            - Case 5: Tabular Section Key Error
                - Evaluate that there is no Tabuler window in "mdi_area"
                - Evaluate that "ErrorDialog" widget is called
            - Case 6:
                - Evaluate that there 3 sub-windows in "mdi_area" when measurement file is loaded
                    - Numeric, Plot, Tabular
        """
        # Setup
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))
        valid_dspf = str(pathlib.Path(self.resource, "valid.dspf"))
        invalid_json_decode_error_dspf = str(pathlib.Path(self.resource, "invalid_JsonDecodeError.dspf"))
        invalid_numeric_section_key_error_dspf = str(pathlib.Path(self.resource, "invalid_NumericSectionKeyError.dspf"))
        invalid_plot_section_key_error_dspf = str(pathlib.Path(self.resource, "invalid_PlotSectionKeyError.dspf"))
        invalid_tabular_section_key_error_dspf = str(pathlib.Path(self.resource, "invalid_TabularSectionKeyError.dspf"))

        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)

        # Case 0
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_0"):
            with (
                mock.patch.object(self.widget, "load_window", wraps=self.widget.load_window) as mo_load_window,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = None, None
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                mo_load_window.assert_not_called()
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

        # Case 1
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_1"):
            with (
                mock.patch.object(self.widget, "load_window", wraps=self.widget.load_window) as mo_load_window,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = valid_dspf[:-2], None
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                mo_load_window.assert_not_called()
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

        # Case 2
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_2"):
            with (
                mock.patch.object(self.widget, "load_window", wraps=self.widget.load_window) as mo_load_window,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = (
                    invalid_json_decode_error_dspf,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                mo_load_window.assert_not_called()
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)
                self.mc_ErrorDialog.assert_called()
                self.mc_ErrorDialog.reset_mock()

        # Case 3
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_3"):
            with (
                mock.patch("asammdf.gui.widgets.file.ErrorDialog") as mc_ErrorDialog,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = (
                    invalid_plot_section_key_error_dspf,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
                mc_ErrorDialog.assert_called()
                mc_ErrorDialog.reset_mock()
                widget_types = self.get_sub_windows()
                self.assertNotIn("Plot", widget_types)

        # Case 4
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_4"):
            with (
                mock.patch("asammdf.gui.widgets.file.ErrorDialog") as mc_ErrorDialog,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = (
                    invalid_numeric_section_key_error_dspf,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
                mc_ErrorDialog.assert_called()
                mc_ErrorDialog.reset_mock()
                widget_types = self.get_sub_windows()
                self.assertNotIn("Numeric", widget_types)

        # Case 5
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_5"):
            with (
                mock.patch("asammdf.gui.widgets.file.ErrorDialog") as mc_ErrorDialog,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = (
                    invalid_tabular_section_key_error_dspf,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
                mc_ErrorDialog.assert_called()
                mc_ErrorDialog.reset_mock()
                widget_types = self.get_sub_windows()
                self.assertNotIn("Tabular", widget_types)

        # Case 6
        with self.subTest("test_PushButton_LoadOfflineWindows_DSPF_6"):
            with (
                mock.patch.object(self.widget, "load_window", wraps=self.widget.load_window) as mo_load_window,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
            ):
                mo_getOpenFileName.return_value = valid_dspf, None
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                mo_load_window.assert_called()
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 3)
                widget_types = self.get_sub_windows()
                self.assertListEqual(["Numeric", "Plot", "Tabular"], widget_types)

        mc_file_ErrorDialog.assert_not_called()

    def test_PushButton_LoadOfflineWindows_LAB(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Ensure that Channels View is set to "Internal file structure"
            - Case 0:
                - Press PushButton: "Load offline windows"
                    - Simulate that "lab" file with empty section was selected
            - Case 1:
                - Press PushButton: "Load offline windows"
                    - Simulate that "lab" with no sections was provided
            - Case 2:
                - Press PushButton: "Load offline windows"
                    - Simulate that valid "lab" file was selected
        Expected:
            - Case 0 & 1:
                - Evaluate that no channel is checked
            - Case 2:
                - Evaluate that 7 channels are checked
        """
        # Setup
        valid_lab = str(pathlib.Path(self.resource, "valid.lab"))
        invalid_missing_section_lab = str(pathlib.Path(self.resource, "invalid_MissingSection.lab"))
        invalid_empty_section_lab = str(pathlib.Path(self.resource, "invalid_EmptySection.lab"))
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))

        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)
        # Switch ComboBox to "Internal file structure"
        self.widget.channel_view.setCurrentText("Internal file structure")
        # Case 0:
        with self.subTest("test_PushButton_LoadOfflineWindows_LAB_0"):
            with mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName:
                mo_getOpenFileName.return_value = (
                    invalid_empty_section_lab,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

                checked_items = []
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    if item.parent() is None:
                        iterator += 1
                        continue
                    if item.checkState(0) == QtCore.Qt.CheckState.Checked:
                        checked_items.append(item.text(0))
                    iterator += 1
                self.assertEqual(0, len(checked_items))
                self.assertNotRegex(
                    str(self.mc_ErrorDialog.mock_calls),
                    r"local variable .* referenced before assignment",
                )

        # Case 1:
        with self.subTest("test_PushButton_LoadOfflineWindows_LAB_1"):
            with mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName:
                mo_getOpenFileName.return_value = (
                    invalid_missing_section_lab,
                    None,
                )
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

                checked_items = []
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    if item.parent() is None:
                        iterator += 1
                        continue
                    if item.checkState(0) == QtCore.Qt.CheckState.Checked:
                        checked_items.append(item.text(0))
                    iterator += 1
                self.assertEqual(0, len(checked_items))
                self.assertNotRegex(
                    str(self.mc_ErrorDialog.mock_calls),
                    r"local variable .* referenced before assignment",
                )

        # Case 2:
        with self.subTest("test_PushButton_LoadOfflineWindows_LAB_2"):
            with (
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getOpenFileName") as mo_getOpenFileName,
                mock.patch("asammdf.gui.widgets.file.QtWidgets.QInputDialog.getItem") as mo_getItem,
            ):
                mo_getOpenFileName.return_value = valid_lab, None
                mo_getItem.return_value = "lab", True
                QtTest.QTest.mouseClick(
                    self.widget.load_channel_list_btn,
                    QtCore.Qt.MouseButton.LeftButton,
                )
                # Evaluate
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

                checked_items = []
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    if item.parent() is None:
                        iterator += 1
                        continue
                    if item.checkState(0) == QtCore.Qt.CheckState.Checked:
                        checked_items.append(item.text(0))
                    iterator += 1
                self.assertEqual(7, len(checked_items))
                self.assertSetEqual(
                    {
                        "ASAM.M.SCALAR.FLOAT64.IDENTICAL",
                        "ASAM.M.SCALAR.UBYTE.FORM_X_PLUS_4",
                        "ASAM_[1].M.MATRIX_DIM_16.UBYTE.IDENTICAL",
                        "ASAM_[11].M.MATRIX_DIM_16.UBYTE.IDENTICAL",
                        "ASAM_[15].M.MATRIX_DIM_16.UBYTE.IDENTICAL",
                        "time",
                    },
                    set(checked_items),
                )

    def test_PushButton_SaveOfflineWindows(self):
        """

        Events:
            - Open 'FileWidget' with valid measurement.
            - Ensure that Channels View is set to "Internal file structure"
            - Press PushButton: "Load offline windows"
                - Simulate that valid "dspf" file was selected
            - Close all Numeric and Tabular windows
            - Add new Plot window with few channels
            - Press PushButton: "Save offline windows"
            - Close all windows.
            - Press PushButton: "Load offline windows"
                - Simulate that saved valid "dspf" file was selected
        Evaluate:
            - Evaluate that new dspf file was saved.
            - Evaluate that two Plot Windows are loaded.
        """
        # Setup
        valid_dspf = str(pathlib.Path(self.resource, "valid.dspf"))
        saved_dspf = pathlib.Path(self.test_workspace, f"{self.id()}.dspf")
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))

        self.setUpFileWidget(measurement_file=measurement_file, default=True)
        # Switch ComboBox to "Internal file structure"
        self.widget.channel_view.setCurrentText("Internal file structure")

        # Load Display File
        self.load_display_file(display_file=valid_dspf)
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 3)

        # Close all Numeric and Tabular windows
        for window in self.widget.mdi_area.subWindowList():
            if window.widget().__class__.__name__ in ("Numeric", "Tabular"):
                window.close()

        # Drag Part

        # Add a new Plot window with few channels
        self.create_window(window_type="Plot", channels_names=["ASAM.M.SCALAR.UBYTE.VTAB_RANGE_NO_DEFAULT_VALUE"])

        # Press PushButton: "Save offline windows"
        with mock.patch("asammdf.gui.widgets.file.QtWidgets.QFileDialog.getSaveFileName") as mo_getSaveFileName:
            mo_getSaveFileName.return_value = str(saved_dspf), None
            QtTest.QTest.mouseClick(
                self.widget.save_channel_list_btn,
                QtCore.Qt.MouseButton.LeftButton,
            )
        # Evaluate
        self.assertTrue(saved_dspf.exists())

        # Event
        self.load_display_file(display_file=saved_dspf)
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
        widget_types = {w.widget().__class__.__name__ for w in self.widget.mdi_area.subWindowList()}
        self.assertSetEqual({"Plot"}, widget_types)

    def test_PushButton_SelectAll(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Switch ComboBox to "Natural sort"
            - Press PushButton: "Select all the channels"
            - Clear selection
            - Switch ComboBox to "Internal file structure"
            - Press PushButton: "Select all the channels"
            - Switch ComboBox to "Selected channels only"
        Evaluate:
            - Evaluate that all channels from "channels_tree" are checked.
        """
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))

        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)

        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural Sort")
        # Press PushButton: "Select all the channels"
        QtTest.QTest.mouseClick(self.widget.select_all_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertTrue(item.checkState(0))
            iterator += 1

        # Clear all
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            item.setCheckState(0, QtCore.Qt.CheckState.Unchecked)
            self.assertEqual(item.checkState(0), QtCore.Qt.CheckState.Unchecked)
            iterator += 1

        # Switch ComboBox to "Internal file structure"
        self.widget.channel_view.setCurrentText("Internal file structure")
        # Press PushButton: "Select all the channels"
        QtTest.QTest.mouseClick(self.widget.select_all_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertTrue(item.checkState(0))
            iterator += 1

        # Switch ComboBox to "Selected channels only"
        self.widget.channel_view.setCurrentText("Selected channels only")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertTrue(item.checkState(0))
            iterator += 1

    def test_PushButton_ClearAll(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Switch ComboBox to "Natural sort"
            - Select all channels
            - Press PushButton: "Clear all selected channels"
            - Switch ComboBox to "Internal file structure"
            - Select all channels
            - Press PushButton: "Clear all selected channels"
            - Switch ComboBox to "Selected channels only"
        Evaluate:
            - Evaluate that all channels from "channels_tree" are unchecked.
        """
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))

        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)

        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural Sort")
        # Select all
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            item.setCheckState(0, QtCore.Qt.CheckState.Checked)
            self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(0))
            iterator += 1
        # Press PushButton: "Clear all selected channels"
        QtTest.QTest.mouseClick(self.widget.clear_channels_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(0))
            iterator += 1

        # Switch ComboBox to "Internal file structure"
        self.widget.channel_view.setCurrentText("Internal file structure")
        while iterator.value():
            item = iterator.value()
            item.setCheckState(0, QtCore.Qt.CheckState.Checked)
            self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(0))
            iterator += 1
        # Press PushButton: "Clear all selected channels"
        QtTest.QTest.mouseClick(self.widget.clear_channels_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(0))
            iterator += 1

        # Switch ComboBox to "Selected channels only"
        self.widget.channel_view.setCurrentText("Selected channels only")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
        while iterator.value():
            item = iterator.value()
            self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(0))
            iterator += 1

    def test_PushButton_Search(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Case 0:
                - Press PushButton: "Search and select channels"
                    - Simulate that "AdvancedSearch" was closed/cancelled (no signal was selected)
            - Case 1:
                - Press PushButton: "Search and select channels"
                    - Simulate that 2 channels were selected and PushButton "Check channels" was pressed.
            - Case 2:
                - Clear 'channel_tree' selection
                - Press PushButton: "Search and select channels"
                    - Simulate that 2 channels were selected and PushButton "Add Window" was pressed.
        Evaluate:
            - Case 0:
                - Evaluate that no channel is selected in "channel_tree"
            - Case 1:
                - Evaluate that 2 channels are selected in "channel_tree"
            - Case 2:
                - Evaluate that 2 channels are selected in "channel_tree"
                - Plot Window is added.
        """
        # Setup
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))
        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)
        # Case 0:
        with self.subTest("test_PushButton_Search_0"):
            with mock.patch("asammdf.gui.widgets.file.AdvancedSearch") as mc_AdvancedSearch:
                mc_AdvancedSearch.return_value.result = {}
                mc_AdvancedSearch.return_value.pattern_window = False
                mc_AdvancedSearch.return_value.add_window_request = False

                # - Press PushButton: "Search and select channels"
                QtTest.QTest.mouseClick(self.widget.advanced_search_btn, QtCore.Qt.MouseButton.LeftButton)
                # Evaluate
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    self.assertEqual(item.checkState(0), QtCore.Qt.CheckState.Unchecked)
                    iterator += 1

        # Case 1:
        with self.subTest("test_PushButton_Search_1"):
            with mock.patch("asammdf.gui.widgets.file.AdvancedSearch") as mc_AdvancedSearch:
                mc_AdvancedSearch.return_value.result = {
                    (4, 3): "ASAM.M.SCALAR.FLOAT64.IDENTICAL",
                    (2, 10): "ASAM.M.SCALAR.FLOAT32.IDENTICAL",
                }
                mc_AdvancedSearch.return_value.pattern_window = False
                mc_AdvancedSearch.return_value.add_window_request = False

                # - Press PushButton: "Search and select channels"
                QtTest.QTest.mouseClick(self.widget.advanced_search_btn, QtCore.Qt.MouseButton.LeftButton)
                # Evaluate
                checked_channels = 0
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    if item.checkState(0) == QtCore.Qt.CheckState.Checked:
                        checked_channels += 1
                    iterator += 1
                self.assertEqual(2, checked_channels)
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

        # Case 2:
        with self.subTest("test_PushButton_Search_2"):
            # - Clear 'channel_tree' selection
            iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
            while iterator.value():
                item = iterator.value()
                item.setCheckState(0, QtCore.Qt.CheckState.Unchecked)
                iterator += 1
            with (
                mock.patch("asammdf.gui.widgets.file.AdvancedSearch") as mc_AdvancedSearch,
                mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog,
            ):
                mc_AdvancedSearch.return_value.result = {
                    (4, 3): "ASAM.M.SCALAR.FLOAT64.IDENTICAL",
                    (2, 10): "ASAM.M.SCALAR.FLOAT32.IDENTICAL",
                }
                mc_AdvancedSearch.return_value.pattern_window = False
                mc_AdvancedSearch.return_value.add_window_request = True
                mc_WindowSelectionDialog.return_value.result.return_value = True
                mc_WindowSelectionDialog.return_value.selected_type.return_value = "New plot window"
                mc_WindowSelectionDialog.return_value.disable_new_channels.return_value = False

                # - Press PushButton: "Search and select channels"
                QtTest.QTest.mouseClick(self.widget.advanced_search_btn, QtCore.Qt.MouseButton.LeftButton)
                # Evaluate
                checked_channels = 0
                iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.channels_tree)
                while iterator.value():
                    item = iterator.value()
                    if item.checkState(0) == QtCore.Qt.CheckState.Checked:
                        checked_channels += 1
                    iterator += 1
                self.assertEqual(2, checked_channels)
                self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
                widget_types = self.get_sub_windows()
                self.assertIn("Plot", widget_types)

    def test_PushButton_CreateWindow(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Switch ComboBox to "Natural sort"
            - Case 0:
                - Press PushButton "Create Window"
                    - Simulate that "WindowSelectionDialog" is cancelled/closed.
            - Case 1:
                - Press PushButton "Create Window"
                    - Simulate that Window Type is Plot.
            - Case 2:
                - Select one channel
                - Press PushButton "Create Window"
                    - Simulate that Window Type is Numeric.
            - Case 3:
                - Press PushButton "Create Window"
                    - Simulate that Window Type is Tabular.
            - Case 4:
                - Press PushButton "Create Window"
                    - Simulate that Window Type is Plot.
        Evaluate:
            - Evaluate that 4 sub-windows are created.
            - Check if windows contain channel selected.
        """
        # Setup
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))
        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)
        self.widget.channel_view.setCurrentText("Natural sort")

        # Case 0:
        with self.subTest("test_PushButton_CreateWindow_0"):
            with mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog:
                mc_WindowSelectionDialog.return_value.result.return_value = False
                # - Press PushButton "Create Window"
                QtTest.QTest.mouseClick(self.widget.create_window_btn, QtCore.Qt.MouseButton.LeftButton)
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 0)

        # Case 1:
        with self.subTest("test_PushButton_CreateWindow_1"):
            with mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog:
                mc_WindowSelectionDialog.return_value.result.return_value = True
                mc_WindowSelectionDialog.return_value.selected_type.return_value = "Plot"
                # - Press PushButton "Create Window"
                QtTest.QTest.mouseClick(self.widget.create_window_btn, QtCore.Qt.MouseButton.LeftButton)
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
            widget_types = self.get_sub_windows()
            self.assertIn("Plot", widget_types)

        # Case 2:
        with self.subTest("test_PushButton_CreateWindow_2"):
            with mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog:
                mc_WindowSelectionDialog.return_value.result.return_value = True
                mc_WindowSelectionDialog.return_value.selected_type.return_value = "Numeric"
                # - Select one channel
                channel = self.widget.channels_tree.topLevelItem(0).text(0)
                self.widget.channels_tree.topLevelItem(0).setCheckState(0, QtCore.Qt.CheckState.Checked)
                # - Press PushButton "Create Window"
                QtTest.QTest.mouseClick(self.widget.create_window_btn, QtCore.Qt.MouseButton.LeftButton)
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
            widget_types = self.get_sub_windows()
            self.assertIn("Numeric", widget_types)
            numeric_data = self.widget.mdi_area.subWindowList()[1].widget().channels.dataView
            numeric_channel = numeric_data.model().data(numeric_data.model().index(0, 0))
            self.assertEqual(channel, numeric_channel)

        # Case 3:
        with self.subTest("test_PushButton_CreateWindow_3"):
            with mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog:
                mc_WindowSelectionDialog.return_value.result.return_value = True
                mc_WindowSelectionDialog.return_value.selected_type.return_value = "Tabular"
                # - Press PushButton "Create Window"
                QtTest.QTest.mouseClick(self.widget.create_window_btn, QtCore.Qt.MouseButton.LeftButton)
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 3)
            widget_types = self.get_sub_windows()
            self.assertIn("Tabular", widget_types)

        # Case 4:
        with self.subTest("test_PushButton_CreateWindow_4"):
            with mock.patch("asammdf.gui.widgets.file.WindowSelectionDialog") as mc_WindowSelectionDialog:
                mc_WindowSelectionDialog.return_value.result.return_value = True
                mc_WindowSelectionDialog.return_value.selected_type.return_value = "Plot"
                # - Press PushButton "Create Window"
                QtTest.QTest.mouseClick(self.widget.create_window_btn, QtCore.Qt.MouseButton.LeftButton)
            # Evaluate
            self.assertEqual(len(self.widget.mdi_area.subWindowList()), 4)
            widget_types = self.get_sub_windows()
            self.assertIn("Plot", widget_types)
            plot_widget = self.widget.mdi_area.subWindowList()[3].widget()
            plot_channel = plot_widget.channel_selection.topLevelItem(0).text(0)
            self.assertEqual(channel, plot_channel)

    def test_DoubleClick_Channel(self):
        """
        Events:
            - Open 'FileWidget' with valid measurement.
            - Ensure that "channel_view" is set to "Internal file structure"
            - Case 0:
                - DoubleClick on Channel Group
            - Case 1:
                - DoubleClick on Channel
        Evaluate:
            - Evaluate that new dialog is visible and display channel meta-data.
        """
        # Setup
        measurement_file = str(pathlib.Path(self.resource, "ASAP2_Demo_V171.mf4"))
        # Event
        self.setUpFileWidget(measurement_file=measurement_file, default=True)
        self.widget.channel_view.setCurrentText("Internal file structure")

        first_item = self.widget.channels_tree.topLevelItem(0)
        first_item_center = self.widget.channels_tree.visualItemRect(first_item).center()
        QtTest.QTest.mouseClick(
            self.widget.channels_tree.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            first_item_center,
        )
        self.processEvents()
        # Case 0:
        with mock.patch(
            "asammdf.gui.widgets.file.ChannelGroupInfoDialog",
            wraps=ChannelGroupInfoDialog,
        ) as mc_ChannelGroupInfoDialog:
            QtTest.QTest.mouseDClick(
                self.widget.channels_tree.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                first_item_center,
            )
            # Evaluate
            mc_ChannelGroupInfoDialog.assert_called()
            # Identify child: ChannelGroupInfoDialog
            for child in self.widget.children():
                if child.__class__.__name__ == "ChannelGroupInfoDialog":
                    break
            if hasattr(child, "isVisible") and hasattr(child, "close"):
                self.assertTrue(child.isVisible())
                child.close()
                self.assertFalse(child.isVisible())

        # Case 1:
        child_item = first_item.child(0)
        child_item_center = self.widget.channels_tree.visualItemRect(child_item).center()
        QtTest.QTest.mouseClick(
            self.widget.channels_tree.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            child_item_center,
        )
        self.processEvents()
        with mock.patch(
            "asammdf.gui.widgets.file.ChannelInfoDialog",
            wraps=ChannelInfoDialog,
        ) as mc_ChannelGroupInfoDialog:
            QtTest.QTest.mouseDClick(
                self.widget.channels_tree.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                child_item_center,
            )
            # Evaluate
            mc_ChannelGroupInfoDialog.assert_called()
            # Identify child: ChannelInfoDialog
            for child in self.widget.children():
                if child.__class__.__name__ == "ChannelInfoDialog":
                    break
            if hasattr(child, "isVisible") and hasattr(child, "close"):
                self.assertTrue(child.isVisible())
                child.close()
                self.assertFalse(child.isVisible())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'numeric_offline.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDoubleSpinBox, QGridLayout,
    QGroupBox, QHBoxLayout, QLabel, QLineEdit,
    QPushButton, QSizePolicy, QSlider, QSpacerItem,
    QVBoxLayout, QWidget)
from . import resource_rc

class Ui_NumericDisplay(object):
    def setupUi(self, NumericDisplay):
        if not NumericDisplay.objectName():
            NumericDisplay.setObjectName(u"NumericDisplay")
        NumericDisplay.resize(681, 666)
        self.main_layout = QVBoxLayout(NumericDisplay)
        self.main_layout.setSpacing(1)
        self.main_layout.setObjectName(u"main_layout")
        self.main_layout.setContentsMargins(1, 1, 1, 1)
        self.controls = QGroupBox(NumericDisplay)
        self.controls.setObjectName(u"controls")
        self.controls.setFlat(True)
        self.verticalLayout = QVBoxLayout(self.controls)
        self.verticalLayout.setSpacing(1)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.gridLayout_2 = QGridLayout()
        self.gridLayout_2.setSpacing(1)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.toggle_controls_btn = QPushButton(self.controls)
        self.toggle_controls_btn.setObjectName(u"toggle_controls_btn")
        icon = QIcon()
        icon.addFile(u":/down.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toggle_controls_btn.setIcon(icon)

        self.gridLayout_2.addWidget(self.toggle_controls_btn, 0, 3, 1, 1)

        self.controls_label = QLabel(self.controls)
        self.controls_label.setObjectName(u"controls_label")

        self.gridLayout_2.addWidget(self.controls_label, 0, 0, 1, 1)

        self.float_precision = QComboBox(self.controls)
        self.float_precision.setObjectName(u"float_precision")

        self.gridLayout_2.addWidget(self.float_precision, 0, 1, 1, 1)

        self.controls_spacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.controls_spacer, 0, 2, 1, 1)

        self.gridLayout_2.setColumnStretch(2, 1)

        self.verticalLayout.addLayout(self.gridLayout_2)

        self.time_group = QGroupBox(self.controls)
        self.time_group.setObjectName(u"time_group")
        self.horizontalLayout_2 = QHBoxLayout(self.time_group)
        self.horizontalLayout_2.setSpacing(1)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(1)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.timestamp = QDoubleSpinBox(self.time_group)
        self.timestamp.setObjectName(u"timestamp")
        self.timestamp.setDecimals(9)
        self.timestamp.setSingleStep(0.001000000000000)

        self.horizontalLayout.addWidget(self.timestamp)

        self.min_t = QLabel(self.time_group)
        self.min_t.setObjectName(u"min_t")

        self.horizontalLayout.addWidget(self.min_t)

        self.timestamp_slider = QSlider(self.time_group)
        self.timestamp_slider.setObjectName(u"timestamp_slider")
        self.timestamp_slider.setMaximum(99999)
        self.timestamp_slider.setOrientation(Qt.Orientation.Horizontal)
        self.timestamp_slider.setInvertedAppearance(False)
        self.timestamp_slider.setInvertedControls(False)
        self.timestamp_slider.setTickPosition(QSlider.TickPosition.NoTicks)
        self.timestamp_slider.setTickInterval(1)

        self.horizontalLayout.addWidget(self.timestamp_slider)

        self.max_t = QLabel(self.time_group)
        self.max_t.setObjectName(u"max_t")

        self.horizontalLayout.addWidget(self.max_t)

        self.horizontalLayout.setStretch(2, 1)

        self.horizontalLayout_2.addLayout(self.horizontalLayout)


        self.verticalLayout.addWidget(self.time_group)

        self.search_group = QGroupBox(self.controls)
        self.search_group.setObjectName(u"search_group")
        self.gridLayout = QGridLayout(self.search_group)
        self.gridLayout.setSpacing(1)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(1, 1, 1, 1)
        self.pattern_match = QLineEdit(self.search_group)
        self.pattern_match.setObjectName(u"pattern_match")

        self.gridLayout.addWidget(self.pattern_match, 0, 1, 1, 1)

        self.op = QComboBox(self.search_group)
        self.op.addItem("")
        self.op.addItem("")
        self.op.addItem("")
        self.op.addItem("")
        self.op.addItem("")
        self.op.addItem("")
        self.op.setObjectName(u"op")

        self.gridLayout.addWidget(self.op, 0, 2, 1, 1)

        self.match_type = QComboBox(self.search_group)
        self.match_type.addItem("")
        self.match_type.addItem("")
        self.match_type.setObjectName(u"match_type")

        self.gridLayout.addWidget(self.match_type, 1, 1, 1, 1)

        self.target = QLineEdit(self.search_group)
        self.target.setObjectName(u"target")

        self.gridLayout.addWidget(self.target, 0, 3, 1, 1)

        self.match_mode = QComboBox(self.search_group)
        self.match_mode.addItem("")
        self.match_mode.addItem("")
        self.match_mode.setObjectName(u"match_mode")

        self.gridLayout.addWidget(self.match_mode, 0, 0, 1, 1)

        self.case_sensitivity = QComboBox(self.search_group)
        self.case_sensitivity.addItem("")
        self.case_sensitivity.addItem("")
        self.case_sensitivity.setObjectName(u"case_sensitivity")

        self.gridLayout.addWidget(self.case_sensitivity, 2, 1, 1, 1)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setSpacing(1)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.backward = QPushButton(self.search_group)
        self.backward.setObjectName(u"backward")
        icon1 = QIcon()
        icon1.addFile(u":/right.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.backward.setIcon(icon1)

        self.horizontalLayout_3.addWidget(self.backward)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_4)

        self.forward = QPushButton(self.search_group)
        self.forward.setObjectName(u"forward")
        icon2 = QIcon()
        icon2.addFile(u":/left.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.forward.setIcon(icon2)

        self.horizontalLayout_3.addWidget(self.forward)


        self.gridLayout.addLayout(self.horizontalLayout_3, 1, 3, 2, 1)

        self.match = QLabel(self.search_group)
        self.match.setObjectName(u"match")

        self.gridLayout.addWidget(self.match, 3, 0, 1, 4)

        self.gridLayout.setColumnStretch(1, 1)

        self.verticalLayout.addWidget(self.search_group)


        self.main_layout.addWidget(self.controls)

        QWidget.setTabOrder(self.timestamp, self.pattern_match)
        QWidget.setTabOrder(self.pattern_match, self.op)
        QWidget.setTabOrder(self.op, self.target)
        QWidget.setTabOrder(self.target, self.backward)
        QWidget.setTabOrder(self.backward, self.forward)

        self.retranslateUi(NumericDisplay)

        QMetaObject.connectSlotsByName(NumericDisplay)
    # setupUi

    def retranslateUi(self, NumericDisplay):
        NumericDisplay.setWindowTitle(QCoreApplication.translate("NumericDisplay", u"Form", None))
        self.controls.setTitle("")
        self.toggle_controls_btn.setText(QCoreApplication.translate("NumericDisplay", u"Show controls", None))
        self.controls_label.setText(QCoreApplication.translate("NumericDisplay", u"Precision", None))
        self.time_group.setTitle(QCoreApplication.translate("NumericDisplay", u"Time stamp", None))
        self.timestamp.setSuffix(QCoreApplication.translate("NumericDisplay", u"s", None))
        self.min_t.setText("")
        self.max_t.setText("")
        self.search_group.setTitle(QCoreApplication.translate("NumericDisplay", u"Search for values", None))
        self.pattern_match.setPlaceholderText(QCoreApplication.translate("NumericDisplay", u"pattern", None))
        self.op.setItemText(0, QCoreApplication.translate("NumericDisplay", u"==", None))
        self.op.setItemText(1, QCoreApplication.translate("NumericDisplay", u"!=", None))
        self.op.setItemText(2, QCoreApplication.translate("NumericDisplay", u"<", None))
        self.op.setItemText(3, QCoreApplication.translate("NumericDisplay", u"<=", None))
        self.op.setItemText(4, QCoreApplication.translate("NumericDisplay", u">", None))
        self.op.setItemText(5, QCoreApplication.translate("NumericDisplay", u">=", None))

        self.match_type.setItemText(0, QCoreApplication.translate("NumericDisplay", u"Wildcard", None))
        self.match_type.setItemText(1, QCoreApplication.translate("NumericDisplay", u"Regex", None))

        self.target.setPlaceholderText(QCoreApplication.translate("NumericDisplay", u"target value", None))
        self.match_mode.setItemText(0, QCoreApplication.translate("NumericDisplay", u"Raw", None))
        self.match_mode.setItemText(1, QCoreApplication.translate("NumericDisplay", u"Scaled", None))

        self.case_sensitivity.setItemText(0, QCoreApplication.translate("NumericDisplay", u"Case insensitive", None))
        self.case_sensitivity.setItemText(1, QCoreApplication.translate("NumericDisplay", u"Case sensitive", None))

        self.backward.setText("")
        self.forward.setText("")
        self.match.setText("")
    # retranslateUi
